/* irsender.cc       letzte Aenderung: 2.7.2012
 Version: 0.0

 Prozessor: ATmega8A mit 16MHz Quarz
 Schaltung: Port B0 mit Frequenzmesser oder Oszilloskop verbunden
                 B1 mit LED
                 B2 mit IR-LED
            Tastenmatrix
		 D2 (INT0) Taste1 (provisorisch)
		 D3 (INT1) Taste2 (provisorisch)
		 fuer 4*5-Matrix:
		 ev. C0-C4 Matrix-Eingaenge
		 ev. D5-D7,C5 Matrix-Ausgaenge

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 1.7.2012	Erstellung

*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include "ulong.h"

#define TAKT_BIT 0  //zur Ausgabe der Modulationsfrequenz fuer Testzwecke 
#define LED1_BIT 1  //erste Kontroll-LED
#define LED2_BIT 3  //zweite Kontroll-LED
#define IRLED_BIT 2 //Ausgabe des Sendesignals

#define NEC_FORMAT  //Format fuer Kaon
#ifdef NEC_FORMAT
#define F_TAKT 38  //Modulations-Frequenz in kHz
#define PULSBREITE 21 //Anzahl Takte pro Puls
#define NULLBITLUECKE 21 //Anzahl Takte fuer 0-Bit
#define EINSBITLUECKE 63 //Anzahl Takte fuer 1-Bit
#define STARTHI 336 //Anzahl Takte fuer Startbit high
#define STARTLO 158 //Anzahl Takte fuer Startbit low
#define STARTLOREP 84 //Anzahl Takte fuer Startbit low im Repeat-Block
#define ADRESSE 0x0410 //Adresse des Geraetes
#define COD_POWER 0x00 //Code fuer Power-Taste
#define COD_1     0x38 //Code fuer 1-Taste
#define COD_2     0xB8 //Code fuer 2-Taste
#define COD_3     0x78 //Code fuer 3-Taste
#define COD_4     0x02 //Code fuer 4-Taste
#define COD_5     0x82 //Code fuer 5-Taste
#define COD_6     0x42 //Code fuer 6-Taste
#define COD_7     0x22 //Code fuer 7-Taste
#define COD_8     0xA2 //Code fuer 8-Taste
#define COD_9     0x62 //Code fuer 9-Taste
#define COD_0     0xE2 //Code fuer 0-Taste
#endif

//provisorisch:
//#define GENAUERUNDUNG
#define TASTE1 PD2
#define TASTE2 PD3

/**** Takterzeugung ****/
static volatile uchar taktzaehler=0,sendeflag=0;
static volatile uchar int0_flag=0,int1_flag=0;

// Genaue Quarzfrequenz in Hertz:
#define F_CPU 16000000L

#ifdef GENAUERUNDUNG
#define NormalerCompwert (((F_CPU/F_TAKT+500)/2000)-1) //bei weniger 0.75 abgerundeter Wert
#define ZweiterCompwert (((F_CPU/F_TAKT+1500)/2000)-1) //bei mehr als 0.25 aufgerundeter Wert
#else
#define NormalerCompwert (((F_CPU/F_TAKT+1000)/2000)-1) //ab 0.5 aufrunden, sonst abrunden
#endif

ISR(SIG_OUTPUT_COMPARE1A)
{
 static uchar takt=0;
 if((takt^=1)!=0)
  {
   taktzaehler++;
#ifdef GENAUERUNDUNG
   OCR1A=NormalerCompwert;
#endif
   PORTB |= (1<<TAKT_BIT); //PB0 setzen
   if(sendeflag!=0) PORTB |= (1<<IRLED_BIT); //falls sendeflag aktiv: IR-LED einschalten
  }
 else
  {
#ifdef GENAUERUNDUNG
   OCR1A=ZweiterCompwert;
#endif
   PORTB &= ~(1<<TAKT_BIT); //PB0 loeschen
   PORTB &= ~(1<<IRLED_BIT); //IR-LED ausschalten
  }
}

void timer1_init()
{
 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1 (also Timer-Takt ist gleich Quarz-Takt)
 OCR1A = NormalerCompwert;
 TCNT1 = 0; //Startwert des Timers
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts bei Vergleichswert
}

//Bei einem unerwarteten Interrupt nichts machen:
void __attribute__ ((naked)) __vector_default(void) {__asm__ ("reti");}

ISR(INT0_vect)
{
 int0_flag=1;
}

ISR(INT1_vect)
{
 int1_flag=1;
}

void ir_startbit()
{
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 PORTB |= (1<<LED2_BIT); //LED2 einschalten
 while(taktzaehler < (STARTHI/2)) ;//erste Haelfte des Startbits abwarten
 taktzaehler=0;
 while(taktzaehler < (STARTHI/2)) ;//zweite Haelfte des Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 PORTB &= ~(1<<LED2_BIT); //LED2 ausschalten
 while(taktzaehler < STARTLO) ;//Low-Teil des Startbits abwarten
}

void ir_bit_senden(uchar data)  //einzelnes Bit ueber IR-Diode senden
{
 uchar luecke;
 sendeflag = 1; //Puls fuer aktuelles Bit senden
 taktzaehler=0;
 PORTB |= (1<<LED2_BIT); //LED2 einschalten
 while(taktzaehler < PULSBREITE) ;//Puls fuer aktuelles Bit abwarten
 sendeflag = 0; //Luecke fuer aktuelles Bit senden
 taktzaehler=0;
 PORTB &= ~(1<<LED2_BIT); //LED2 ausschalten
 luecke = (data==0) ? NULLBITLUECKE : EINSBITLUECKE;
 while(taktzaehler < luecke) ;//Luecke fuer aktuelles Bit abwarten
}

void ir_byte_senden(uchar data)  //ein Byte ueber IR-Diode senden
{
 uchar maske;
 for(maske=0x80;maske!=0;maske>>=1)
  {
   ir_bit_senden(data&maske);
  }
}

void ir_repeat()
{
 //Hier sollte noch ein repeat-code gesendet werden,
 //ist aber fuer Kaon nicht unbedingt noetig.
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 PORTB |= (1<<LED2_BIT); //LED2 einschalten
 while(taktzaehler < (STARTHI/2)) ;//erste Haelfte des Startbits abwarten
 taktzaehler=0;
 while(taktzaehler < (STARTHI/2)) ;//zweite Haelfte des Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 PORTB &= ~(1<<LED2_BIT); //LED2 ausschalten
 while(taktzaehler < STARTLOREP) ;//Low-Teil des Startbits abwarten
 ir_bit_senden(1);
}

void milliwait(int n)
{
 taktzaehler=0;
 while(n>0)
  {
   while(taktzaehler<F_TAKT) ;//eine Millisekunde abwarten
   taktzaehler=0;
   --n;
  }
}

int main(void)
{
 uchar n=0;
 DDRB  = 0x0F; //PortB PB0-3 als Ausgaenge
 PORTB = n; //alle Auasgaenge aus
 //DDRD = 0; //alles Eingaenge
 PORTD = (1<<PD2) | (1<<PD3); //Pullup-Widerstaende setzen

 //Interrupts vorbereiten:
 MCUCR |= (1<<ISC01);//INT0 bei fallender Flanke
 MCUCR |= (1<<ISC11);//INT1 bei fallender Flanke
 GICR |= (1<<INT0);//INT0-Interrupts erlauben
 GICR |= (1<<INT1);//INT1-Interrupts erlauben
 timer1_init();
 sei(); //Interrupt einschalten

 while(1) //Hauptschlaufe
  {
   if(int0_flag!=0 || int1_flag!=0)
    {
     //Entprellung:
     milliwait(1);
     if(int0_flag!=0 && (PIND & (1<<TASTE1))!=0) int0_flag=0; //wenn Taste schon wieder losgelassen --> Stoerpuls, ignorieren
     if(int1_flag!=0 && (PIND & (1<<TASTE2))!=0) int1_flag=0; //wenn Taste schon wieder losgelassen --> Stoerpuls, ignorieren

     if(int0_flag!=0)
      {
       ir_startbit();
       ir_byte_senden(ADRESSE>>8);//Geraeteadresse senden
       ir_byte_senden(ADRESSE&0xFF);//Geraeteadresse senden
       ir_byte_senden(COD_POWER);//Code fuer Powertaste senden
       ir_byte_senden(COD_POWER^0xFF);//Inverser Code fuer Powertaste senden
       ir_bit_senden(0);//abschliessendes Bit senden
       do
	{
	 milliwait(200);
	 ir_repeat();
	}
       while((PIND & (1<<TASTE1))==0);
       milliwait(50);
       taktzaehler=0;
       int0_flag=0;
      }

     if(int1_flag!=0)
      {
       ir_startbit();
       ir_byte_senden(ADRESSE>>8);//Geraeteadresse senden
       ir_byte_senden(ADRESSE&0xFF);//Geraeteadresse senden
       ir_byte_senden(COD_1);//Code fuer 1-Taste senden
       ir_byte_senden(COD_1^0xFF);//Inverser Code fuer 1-Taste senden
       ir_bit_senden(0);//abschliessendes Bit senden
       do
	{
	 milliwait(200);
	 ir_repeat();
	}
       while((PIND & (1<<TASTE2))==0);
       milliwait(50);
       taktzaehler=0;
       int1_flag=0;
      }

    }
   else if(taktzaehler>=38/2) //test: LED1 mit Blinkfrequenz 38 mal kleiner als Takt --> 1kHz
    {
     //cli(); taktzaehler -= 38/2; sei();
     taktzaehler=0;
     //LED1 blinken lassen:
     n ^= 2;
     if(n&2) PORTB |= (1<<LED1_BIT);
     else    PORTB &= ~(1<<LED1_BIT);
    }
  }

 return 0;
}
