;-----------------------------------------------------------------------------
;* irempf13a.asm
;* Version: 0.0 	letzte Aenderung: 29.7.2012
;* Prozessor: ATtiny13A, interner Oszillator 1MHz
;* Schaltschema:
;*   irempf13a.png
;*
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 29.7.2012	Erstellung
;*
;*
;* globale Registerbelegung:
;* r10...r15 Zeiten (1msec bis Tage)
;*
;* global im Hauptprogramm:
;* r9 = Status
;*
;* Verwendung:
;*

;.equ FREQTEST=1
;.equ TIMERTEST=1
;.equ LEDTEST=1

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
	reti			; 3   PCINT0
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	reti			; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	reti			; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ richtung=SRAMSTART		;Motoransteuerung (0b01 oder 0b10)
.equ sensor1adwertL=SRAMSTART+1
.equ sensor1adwertH=SRAMSTART+2 ;Wert von AD-Wandler fuer Sensor

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
        out SPL, r16  	     ; Init Stackpointer

; Prescaler 8 sollte schon mit den Fuses gesetzt sein:
;	ldi r16, 0x80
;	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
;	ldi r16, 3
;	out CLKPR, r16		; Clk-Prescaler auf 8 setzen --> 1MHz

	ldi r16, 0x65		; Wert ausprobieren fuer 1MHz
	out OSCCAL, r16		; Oszillator kallibrieren --> 1MHz

	ldi r16, 0x03		;PB0, PB1 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0x18		;Pullupwiderstaende PB4,PB3 fuer Taster und TSOP
	out PORTB, r16		;LEDs aus, Pullup-Widerstaende setzen

.ifdef FREQTEST
T1:	inc r16			; Test fuer Frequenzmessung
	out PORTB, r16		; Schlaufe braucht 4 Takte, eine Periode also 8
	rjmp T1			; Bei 1MHz muesste an B0 125kHz zu messen sein.
.endif

;; Timer auf 125kHz (8usec) und 125 setzen: 125*8usec=1msec
	ldi r17, 0x02		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
	ldi r17, 0		;0 = Normal-Modus  (2 = CTC-Modus)
	out TCCR0A, r17		;Timer-Modus setzen
	ldi r17, 2		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 8 setzen (1MHz/8=125kHz)
.equ	TCNTSTART=256-125
	ldi r17, TCNTSTART	;Startwert zum Teilen durch 125
	out TCNT0, r17		;Counter auf Startwert setzen

	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
;	clr r15			; Tage, auf 0 setzen
	sei			;Interrupt erlauben

.ifdef TIMERTEST
T1:	nop
	out PORTB, r12		; Sekunden auf LEDs
	rjmp T1
.endif

;-----------------------------------------------------------------------------
	clr r9			; Ausgangs-Status loeschen
;	in r16, PINB		; Eingaenge lesen
;	lsl r16
;	swap r16		; PB4,PB3 auf Bits 1,0 bringen
;	andi r16, 0x03		; restliche Bits loeschen

;; normaler Hauptloop
mainloop:
.ifdef LEDEST
;; Test: LED jede Sekunde kurz aufleuchten lassen:
	mov r16, r12
_m3:	cp r16, r12
	breq _m3		; auf neue Sekunde warten
	sbi PORTB, 1		; rote LED ein
	rcall waitfuenftelsec
	cbi PORTB, 1		; rote LED aus
.else
	ldi r16, 1
	clr r17
	rcall milliwait
.endif

;; Ausgang ein/aus-schalten mit Taster:
	in r16, PINB		; Eingaenge lesen
	andi r16, 0x10		; PB4 testen
	brne ircheck		; gesetzt --> Taste nicht gedrueckt
status_wechseln:
	inc r9			; Ausgangs-Status wechseln
	mov r16, r9
	andi r16, 1
	breq ausgang_aus
ausgang_ein:
	sbi PORTB, 0		; Ausgang ein
	rjmp blink
ausgang_aus:
	cbi PORTB, 0		; Ausgang aus
blink:	sbi PORTB, 1		; rote LED ein
	rcall waitfuenftelsec
	cbi PORTB, 1		; rote LED aus
	rcall waitfuenftelsec
	rjmp mainloop

;; Ausgang ein/aus-schalten mit IR-Fernsteuerung:
ircheck:
	in r16, PINB		; Eingaenge lesen
	andi r16, 0x08		; PB3 testen
	brne irno		; gesetzt --> kein IR-Empfang
;; provisorisch: hier muesste noch der empfangene Code geprueft werden
	rjmp status_wechseln	; bei Erfolg --> ein/aus-schalten	
irno:	rjmp mainloop		; sonst weiter mit Hauptschlaufe
	
;stopp:	cli			; Interrupts aus
;	ldi r16, 0		; alle LEDs aus
;	out PORTB, r16
;	out DDRB, r16		; alle Ports auf Eingang
;	ldi r16, (1<<SE)|(1<<SM1) ;Power-down setzen
;	out MCUCR, r16
;	sleep
;	rjmp main		;sollte nie erreicht werden.


;div18:		;r17:r16 /= 2^r18 ,r18 zerstoert
;	asr r17
;	ror r16
;	dec r18
;	brne div18
;	ret

waitfuenftelsec:
	push r17
	ldi r16, low(200)
	ldi r17, high(200)
	rcall milliwait		; 0.2 Sec
	pop r17
	ret

;wait2einhalb:			; 2.5 sec warten, r16r17 zerstoert
;	ldi r16, low(500)
;	ldi r17, high(500)
;	rcall milliwait		; 0.5 Sec
;	ldi r16, 2		; 2 Sec
;	rjmp secwait

;secwait:	; r16 Sekunden lang warten
;	push r17
;	push r18
;	mov r18, r16
;	add r18, r16
;_sw1:	ldi r16, low(500)
;	ldi r17, high(500)
;	rcall milliwait
;	dec r18
;	brne _sw1
;	pop r18
;	pop r17
;	ret

;-----------------------------------------------------------------------------
; OVR-Interrupt-Routine, wird alle Millisekunden aufgerufen
; Registerbelegungen:
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	in r16, SREG	;Statusregister sichern
	push r16
	push r17
millisekzaehlen:
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
;	ldi r16, 24
;	cp r14, r16
;	brne end1	;fertig wenn std!=24
;	clr r14		;Stunden auf 0 setzen
;	inc r15		;Tage erhoehen
end1:
	ldi r17, TCNTSTART	;fuer Normal-Modus
	out TCNT0, r17		;Counter auf Startwert setzen
	pop r17
	pop r16
	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti

;-----------------------------------------------------------------------------
milliwait:			;Wartezeit: r17:r16 msec, min 1 max 999
	push r18
	push r19
	push r16
	push r17
	cli
	mov r18, r10
	mov r19, r11		;r19r18 = sollzeit = aktuelle Millisec
	sei
	add r18, r16
	adc r19, r17		;sollzeit += Wartezeit
	ldi r16, 0xE8		;auf 1000 testen
	cpi r19, 0x03		;oberes Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
	brne subtra		;sollzeit>1000 --> subtrahiere 1000
	cp r18, r16		;unteres Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
subtra:	sub r18, r16
	sbci r19, 0x03		;wenn sollzeit>=1000 dann sollzeit-=1000
loop1:	cp r18, r10
	brne loop1
	cp r19, r11		;auf sollzeit warten
	brne loop1
	pop r17
	pop r16
pop1918ret:
	pop r19
	pop r18
	ret

;-----------------------------------------------------------------------------
microwait:      ;; genau r16 usec warten, bei 1MHz Takt
		;; mindestens 22, maximal 255
	subi r16, 18		; Anzahl Takte ausserhalb der Schlaufe subtrahieren
	push r16
	andi r16, 3		; Rest beim Dividieren durch 4
	breq mic_r0		; Rest==0 --> kein zusaetzlicher Takt
	dec r16
	breq mic_r1		; Rest==1 --> 1 zusaetzlicher Takt
	dec r16
	breq mic_r2		; Rest==2 --> 2 zusaetzliche Takte
	rjmp mic_r3		; sonst Rest==3 --> 3 zusaetzliche Takte
mic_r0:	nop
mic_r1:	nop
mic_r2:
mic_r3:	pop r16
	lsr r16
	lsr r16			; r16 /= 4
mic_L4:	nop
	dec r16
	brne mic_L4		; Schlaufe braucht 4usec
	ret

;-----------------------------------------------------------------------------
.ifdef TEMPSENSOR
tabelle:	;static int tabelle[]=
	;; Temparaturen 10*mehr damit wir keine Fliesszahlen brauchen
	;; -550 ist also -55.0 GradCelsius
	.dw -550, 990, -500, 1040, -400, 1146, -300, 1260, -200, 1381
	.dw -100, 1510, 0, 1646, 100, 1790, 200, 1941, 250, 2020
	.dw 300, 2100, 400, 2267, 500, 2441, 600, 2623, 700, 2812
	.dw 800, 3009, 900, 3214, 1000, 3426, 1100, 3643, 1200, 3855
	.dw 1250, 3955, 1300, 4048, 1400, 4208, 1500, 4323, 0, 0
.endif

;-----------------------------------------------------------------------------
; Multiplikationen und Divisionen
;
.ifdef MULU16BENOETIGT
mulu16:		;r3:r2:r1:r0 = r1:r0 * r3:r2, Vorzeichenlose Multiplikation
	push r4
	push r5
	push r16
	push r17
	push r18
	push r19
	mov r4, r2
	mov r5, r3	;Multiplikator = r5,r4 = r3,r2
	mov r16, r0
	mov r17, r1
	clr r18
	clr r19		;Multiplikant = r19,r18,r17,r16 = r1,r0
	clr r0
	clr r1
	clr r2
	clr r3		;resultat = r3,r2,r1,r0 = 0
_mu1:	lsr r5
	ror r4		;niedrigstes Bit aus Multiplikator lesen
	brcc _mu2
	add r0, r16	;wenn gesetzt: Multiplikant zum Resultat addieren
	adc r1, r17
	adc r2, r18
	adc r3, r19
_mu2:	lsl r16
	rol r17
	rol r18
	rol r19		;Multiplikant verdoppeln
	tst r4		;noch weitere Bits in Multiplikator vorhanden?
	brne _mu1	;ja -->
	tst r5
	brne _mu1	;ja -->
	pop r19
	pop r18
	pop r17
	pop r16
	pop r5
	pop r4
	ret
.endif
mul16:		  ;r3:r2:r1:r0 = r1:r0 * r3:r2, Multiplikation mit Vorzeichen
	push r18	;(neu geschrieben weil mul im Tiny13 nicht vorhanden)
	push r19
	push r24
	push r25
	push r26
	push r27
	mov r24, r2
	mov r25, r3	;Multiplikator = r25,r24 = r3,r2
	mov r26, r0
	mov r27, r1
	clr r18
	clr r19		;Multiplikant = r19,r18,r27,r26 = r1,r0
	clr r0
	clr r1
	clr r2
	clr r3		;resultat = r3,r2,r1,r0 = 0
	sbrs r25, 7	;ist Multiplikator negativ?
	rjmp _mu0	;nein -->
	com r24		;ja: Negieren
	com r25
	adiw r25:r24, 1
	com r26		;und Multiplikant auch negieren
	com r27
	adiw r27:r26, 1
_mu0:	sbrs r27, 7	;ist Multiplikant negativ?
	rjmp _mu1	;nein -->
	ldi r18, 0xFF	;ja: obere Bits setzen
	ldi r19, 0xFF	
_mu1:	lsr r25
	ror r24		;niedrigstes Bit aus Multiplikator lesen
	brcc _mu2
	add r0, r26	;wenn gesetzt: Multiplikant zum Resultat addieren
	adc r1, r27
	adc r2, r18
	adc r3, r19
_mu2:	lsl r26
	rol r27
	rol r18
	rol r19		;Multiplikant verdoppeln
	tst r24		;noch weitere Bits in Multiplikator vorhanden?
	brne _mu1	;ja -->
	tst r25
	brne _mu1	;ja -->
	pop r27
	pop r26
	pop r25
	pop r24
	rjmp pop1918ret

div16:			;Division wie divu16 aber mit Vorzeichen
	push r4			; r4 = Vorzeichen
	mov r4, r3
	eor r4, r17		; oberstes Bit von r4 ist Vorzeichen
	sbrc r3, 7		;ist zu dividierende Zahl negativ?
	rcall negierer3bisr0	;ja: negieren
	sbrs r17, 7		;ist Divisor negativ?
	rjmp di_1		;nein -->
	com r16			;ja: Divisor negieren
	com r17
	subi r16, -1
	sbci r17, -1
di_1:	rcall divu16
	rol r4			;ist Vorzeichen gesetzt?
	pop r4
	brcc di_ret		; nein -->
				;ja: Resultat negieren
negierer3bisr0:
	com r0
	com r1
	com r2
	com r3		;Einercomplement und Cy setzen
	inc r0
	brne di_ret
	adc r1, r0	;da r0=0 ist und Cy vom com gesetzt, wird 1 addiert
	adc r2, r0
	adc r3, r0
di_ret:	ret

divu16:			;r3:r2:r1:r0 /= r17:r16, r17:r16 = Rest
	push r18	;(von thermologger.asm kopiert)
	push r19
	push r20
	clr r18
	clr r19			;z = r19:r18 = 0
	ldi r20, 32		;32 Bits behandeln
div16L1:lsl r0
	rol r1
	rol r2
	rol r3
	rol r18
	rol r19			;zu dividierende Zahl ins z schieben
	cp  r18, r16
	cpc r19, r17
	brcs div16L2		;ist z groesser oder gleich Divisor?
	sub r18, r16		;ja: z -= Divisor
	sbc r19, r17
	inc r0			;und aktuelles Bit im Resultat setzen
div16L2:dec r20
	brne div16L1		;die Schlaufe 32 mal machen
	mov r16, r18
	mov r17, r19		;r17:r16 = Rest
	pop r20
;	pop r19
;	pop r18
;	ret
	rjmp pop1918ret		;Optimiert weil Speicher im Tiny13 knapp wird.
