;***************************************************************************
;*
;* "div16u" - 16/16 Bit Unsigned Division
;*
;* This subroutine divides the two 16-bit numbers 
;* "dd8uH:dd8uL" (dividend) and "dv16uH:dv16uL" (divisor). 
;* The result is placed in "dres16uH:dres16uL" and the remainder in
;* "drem16uH:drem16uL".
;*  
;* Number of words	:16 (was 19 in original)
;* Number of cycles	:178/210 (Min/Max) (was 235/251 in original)
;* Low registers used	:2 (drem16uL,drem16uH)
;* High registers used  :5 (dres16uL/dd16uL,dres16uH/dd16uH,dv16uL,dv16uH,
;*			    dcnt16u)
;*
;***************************************************************************

;***** Subroutine Register Variables

.def	drem16uL=r14
.def	drem16uH=r15
.def	dres16uL=r16
.def	dres16uH=r17
.def	dd16uL	=r16
.def	dd16uH	=r17
.def	dv16uL	=r18
.def	dv16uH	=r19
.def	dcnt16u	=r20

;***** Code

div16u:	clr	drem16uL	;clear remainder Low byte
	sub	drem16uH,drem16uH;clear remainder High byte and carry
	ldi	dcnt16u,16	;init loop counter
d16u_1:	lsl	dd16uL		;shift left dividend
	rol	dd16uH
	rol	drem16uL	;shift dividend into remainder
	rol	drem16uH
	cp	drem16uL,dv16uL	;check remainder - divisor
	cpc	drem16uH,dv16uH	;
	brcs	d16u_3		;if result not negative
	sub	drem16uL,dv16uL	;    remainder = remainder - divisor
	sbc	drem16uH,dv16uH
	inc	dd16uL		;    set the bit shifted into result
d16u_3:	dec	dcnt16u		;decrement counter
	brne	d16u_1		;if done
	ret			;    return
