;-----------------------------------------------------------------------------
;* irempf13a.asm
;* Version: 0.3 	letzte Aenderung: 22.2.2014
;* Prozessor: ATtiny13A, interner Oszillator 1MHz
;* Schaltschema:
;*   irempf13a.png
;*
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 29.7.2012	Erstellung
;* 4.8.12       Versuch PCINT3 zu verwenden
;* 23.3.2013	Fehler beim PCINT3 einstellen korrigiert: out GIMSK, r16 (nicht sts)
;* 18.2.2014	Versuch zu schalten bei Erkennung von Samsung-Startcode
;* 22.2.2014	Gekuerzt auf wirklich gebrauchte Teile, und vereinfacht
;*
;*
;* globale Registerbelegung:
;* r10...r15 Zeiten (1msec bis Tage) fuer Uhr
;* r9 8tel Millisec fuer Uhr
;* r0 = SREG speichern im Timer-Interrupt
;* r3 = TCNTSTART (mit CTC-Modus nicht noetig)
;* r8 = Erkennung von IR-Signal (setzen in Interrupt-Routine, im Hauptprogramm ruecksetzen)
;* r25:r24 Zeitdifferenz zwischen Flanken
;* r23:r22 empfangene IR-Bits
;* r20 = Bitzaehler
;* r21 = Geraet-Erkennung, 0=warten auf Startbit
;*
;* global im Hauptprogramm:
;* r7 = Status
;*

;.equ FREQTEST=1		;Austesten der CPU-Frequenz
;.equ TIMERTEST=1		;Austesten des Timers
.equ TIMER_CTC_MODUS=1		;genauere Version des Timers
;.equ PCINT3_VERWENDEN=1

;; Definition der Geraete (Formate):
.equ SAMSUNG=1
.equ KAON=2
.equ PANASON=3
.equ SCHALTGERAET=SAMSUNG

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
.ifdef PCINT3_VERWENDEN
	rjmp	pc_int0		; 3   PCINT0
.else
	reti			; 3   PCINT0
.endif
.ifdef TIMER_CTC_MODUS
	reti			; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	rjmp	tc0compa	; 7   Timer0 CompareA
.else
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
.endif
	reti			; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	reti			; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
        out SPL, r16  	     ; Init Stackpointer
	ldi r16, 0x03		;PB0, PB1 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, (1<<4)|(1<<3)	;Pullupwiderstaende PB4,PB3 fuer Taster und TSOP
	out PORTB, r16		;LEDs aus, Pullup-Widerstaende setzen

; Prescaler 8 sollte schon mit den Fuses gesetzt sein:
;	ldi r16, 0x80
;	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
;	ldi r16, 3
;	out CLKPR, r16		; Clk-Prescaler auf 8 setzen --> 1MHz

	ldi r16, 0x62		; Wert ausprobieren fuer 1MHz
	out OSCCAL, r16		; Oszillator kallibrieren --> 1MHz

.ifdef FREQTEST
T1:	inc r16			; Test fuer Frequenzmessung
	out PORTB, r16		; Schlaufe braucht 4 Takte, eine Periode also 8
	rjmp T1			; Bei 1MHz muesste an B0 125kHz zu messen sein.
.endif

;; Timer Takt auf 1MHz und Teiler auf 125 setzen damit ein Interrupt alle 0.125msec (8kHz) erfolgt.
.ifdef TIMER_CTC_MODUS
	ldi r17, 0x04		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
	ldi r17, 2		;0 = Normal-Modus  (2 = CTC-Modus)
	out TCCR0A, r17		;Timer-Modus setzen
;; Vorteiler im TCCR0B: 1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
;	ldi r17, 2		;bei F_CPU==8MHz: Vorteiler auf 8 setzen (8MHz/8=1MHz)
	ldi r17, 1		;bei F_CPU==1MHz: nicht teilen
	out TCCR0B, r17
	ldi r17, 125-1
	out OCR0A, r17		;Vergleichsregister A setzen
.else
	ldi r17, 0x02		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
	ldi r17, 0		;0 = Normal-Modus  (2 = CTC-Modus)
	out TCCR0A, r17		;Timer-Modus setzen
;; Vorteiler im TCCR0B: 1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
;	ldi r17, 2		;bei F_CPU==8MHz: Vorteiler auf 8 setzen (8MHz/8=1MHz)
	ldi r17, 1		;bei F_CPU==1MHz: nicht teilen
	out TCCR0B, r17
.equ	TCNTSTART=256-125
	ldi r17, TCNTSTART	;Startwert zum Teilen durch 125
	out TCNT0, r17		;Counter auf Startwert setzen
	subi r17, -4		;Um Taktzyklen von Interrupt-Aufruf korrigierter Wert
	mov r3, r17		;in r3 speichern.
.endif

	ldi r16, 8
	mov r9, r16		;8tel Millisec auf Startwert setzen
	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
;	clr r14			; Stunden
;	clr r15			; Tage, auf 0 setzen

;; Interrupt fuer Aenderungen auf PB3 vorbereiten:
.ifdef PCINT3_VERWENDEN
	sbi PCMSK, PCINT3
	ldi r16, (1<<PCIE)
	out GIMSK, r16
.endif

	sei			;Interrupts erlauben

.ifdef TIMERTEST
T1:	out PORTB, r12		;Sekunden anzeigen
	tst r13
	breq T1
T2:	out PORTB, r13		;Minuten anzeigen
	rjmp T2
.endif

;-----------------------------------------------------------------------------
	clr r21			; Geraete-Erkennung auf 0 (warte auf Startbit)
	clr r7			; Ausgangs-Status loeschen (zu schaltender Ausang aus)
	clr r8			; Flag fuer IR-Signal loeschen

;; normaler Hauptloop
mainloop:

;; Ausgang ein/aus-schalten mit Taster:
	sbis PINB, 4		; PB4 testen (skip bit ioport set)
	rcall status_wechseln	; geloescht --> Taste gedrueckt

;; Ausgang ein/aus-schalten mit IR-Fernsteuerung:
	tst r8			; IR-Signal von Interruptroutine erkannt?
	brne ir_erkannt		; ja -->

.ifdef PCINT3_VERWENDEN
	;; hier koennte man noch sleep einbauen zum Strom sparen
.else
	sbis PINB, 3		; PB3 testen
	rcall pc_int0		; auf Low: Interruptroutine aufrufen
.endif
	rjmp mainloop

ir_erkannt:
	mov r16, r8
	cpi r16, SCHALTGERAET	;richtiges Geraet erkannt?
	brne ir_L1		;nein-->
	;; hier koennte noch nach Nummer (r23:r22) abgefragt werden
	rcall status_wechseln
ir_L1:
	mov r17, r8		;test: rote LED dem Geraet entsprechnde Anzahle mal blinken lassen
	cpi r17, SCHALTGERAET
	breq tst2		;einmal wurde schon geblinkt
tst1:	rcall blink
tst2:	dec r17
	brne tst1
	clr r8			; IR-Status fuer Interruptroutine ruecksetzen
	rjmp mainloop

status_wechseln:
	inc r7			; Ausgangs-Status wechseln
	mov r16, r7
	andi r16, 1
	breq ausgang_aus
ausgang_ein:
	sbi PORTB, 0		; Ausgang ein (set bit ioport)
	rjmp blink
ausgang_aus:
	cbi PORTB, 0		; Ausgang aus (clear bit ioport)
blink:	sbi PORTB, 1		; rote LED ein
	rcall waitfuenftelsec
	cbi PORTB, 1		; rote LED aus
	rcall waitfuenftelsec
	ret

waitfuenftelsec:
	push r17
	ldi r16, low(200)
	ldi r17, high(200)
	rcall milliwait		; 0.2 Sec
	pop r17
	ret

;-----------------------------------------------------------------------------
; Timer-Interrupt-Routine, wird alle 0.125 Millisekunden aufgerufen
; Registerbelegungen:
; r9 8tel Millisekunden, r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0compa:
tc0over:
.ifdef TIMER_CTC_MODUS
.else
	out TCNT0, r3	;Counter auf Startwert (TCNTSTART) setzen
.endif
	push r16
	in r0, SREG	;Statusregister sichern
	adiw r25:r24, 1	;Zeitdifferenzen in 1/8-Microsekunden
	brne L0
	ldi r25, 0xFF	;bei Ueberlauf r25:r24 auf hohen Wert setzen
L0:	dec r9
	brne end2
	ldi r16, 8
	mov r9, r16
millisekzaehlen:
	push r17
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
;	cp r13, r16
;	brne end1	;fertig wenn min!=60
;	clr r13		;Minuten auf 0 setzen
;	inc r14		;Stunden erhoehen
;	ldi r16, 24
;	cp r14, r16
;	brne end1	;fertig wenn std!=24
;	clr r14		;Stunden auf 0 setzen
;	inc r15		;Tage erhoehen
end1:	pop r17
end2:	out SREG, r0	;Statusregister zurueckholen
	pop r16
	reti

;-----------------------------------------------------------------------------
; PCINT0-Interrupt-Routine, bei Aenderungen von PB3 (PCINT3) aufgerufen
; (bei Version mit Atmega8 bei Aenderung von PC3 vom Hauptprogramm aus aufgerufen)

; Zeiten in 1/8 Millisekunden definieren:
.equ SAMSUNG_STARTL=36		;Startbit 4.5ms auf Low
.equ SAMSUNG_STARTH=36		;Startbit 4.5ms auf High
.equ KAON_STARTL=72		;Startbit 9ms auf Low
.equ KAON_STARTH=36		;Startbit 4.5ms auf High
.equ PANASON_STARTL=28		;Startbit 3.5ms auf Low
.equ PANASON_STARTH=14		;Startbit 1.75ms auf High

pc_int0:
	push r16
	in r16, SREG		;Statusregister sichern
	tst r8			;Signal schon erkannt?
	brne pc_ret		;ja --> nichts tun
	sbic PINB, 3		;PB3 auf low?
	rjmp pc_ret		;nein --> nichts tun (nur fallende Flanke auswerten)
	sei			;Interrupts fuer den Timer wieder freigeben
	push r16
	push r17
	clr r24			;r25:24 ruecksetzen
	clr r25
;	rjmp pc_code_erkannt	;test
	cpi r21, SAMSUNG	;SAMSUNG erkannt?
	breq pc_samsung		;ja --> Samsung-Code weiter auswerten
	cpi r21, KAON		;KAON erkannt?
	breq pc_kaon		;ja --> Kaon-Code weiter auswerten
	cpi r21, PANASON	;PANASON erkannt?
	breq pc_panason		;ja --> Panasonic-Code weiter auswerten
	;; hier koennten noch weitere Geraete (Formate) eingefuegt werden
pc_startbit:			;Laenge des Startbits messen:
pc_w1:	sbis PINB, 3		; PB3 auf high?
	rjmp pc_w1		; nein-> warten bis high
	tst r25
	brne startbit_falsch	  ;Starbit in allen Faellen zu lang
	cpi r24, SAMSUNG_STARTL-3 ;kleiner als 0.25 ms weniger als erwartete Startbitlaenge?
	brcs pc_g1		  ;ja-> Startbit zu kurz fuer Samsung
	cpi r24, SAMSUNG_STARTL+3 ;Startbitlaenge +- 0.25 ms von erwarteter fuer SAMSUNG-Code?
	brcs samsung_startbit	  ;ja-> Startbit erkannt
pc_g1:	cpi r24, KAON_STARTL-4 	  ;kleiner als 0.5 ms weniger als erwartet fuer KAON?
	brcs pc_g2		  ;ja-> Startbit zu kurz fuer Kaon
	cpi r24, KAON_STARTL+4 	  ;Startbitlaenge +- 0.5 ms von erwarteter fuer KAON-Code?
	brcs kaon_startbit	  ;ja-> Startbit erkannt
pc_g2:	cpi r24, PANASON_STARTL-3 ;kleiner als 0.25 ms weniger als erwartet fuer KAON?
	brcs pc_g3		  ;ja-> Startbit zu kurz fuer Panason
	cpi r24, PANASON_STARTL+3 ;Startbitlaenge +- 0.25 ms von erwarteter fuer KAON-Code?
	brcs panason_startbit	  ;ja-> Startbit erkannt
pc_g3:
	;; hier koennten noch weitere Geraete (Formate) eingefuegt werden
startbit_falsch:
	ldi r16, low(300)
	ldi r17, high(300)
	rcall milliwait		;Warten bis aktuelle Pulse vorbei
	clr r21
	rjmp pc_int0_fertig
samsung_startbit:
	ldi r21, SAMSUNG
	clr r20			;Bitzaehler ruecksetzen
	rjmp pc_int0_fertig
pc_samsung:
	nop			;hier noch programmieren zum SAMSUNG-Code weiter auswerten
	rjmp pc_code_erkannt
kaon_startbit:
	ldi r21, KAON
	clr r20			;Bitzaehler ruecksetzen
	rjmp pc_int0_fertig
pc_kaon:
	nop			;hier noch programmieren zum KAON-Code weiter auswerten
	rjmp pc_code_erkannt
panason_startbit:
	ldi r21, PANASON
	clr r20			;Bitzaehler ruecksetzen
	rjmp pc_int0_fertig
pc_panason:
	nop			;hier noch programmieren zum PANASON-Code weiter auswerten
pc_code_erkannt:
	mov r8, r21	; Flag fuer erfolgreichen IR-Empfang setzen (auf Geraetewert)
	clr r21		; Status wieder auf 0 fuer naechste Erkennung
pc_int0_fertig:
	pop r17
	pop r16
pc_ret:	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti

;-----------------------------------------------------------------------------
;; Routinen fuer Aufruf aus pc_int0
bit_speichern:
	clc			; Cy loeschen
	sbis PINB, 3		; PB3 gesetzt?
	sec			; nein: Cy setzen
	rol r22
	rol r23			; Neues Bit in r23:r22 schieben
	inc r20			; Bits zaehlen
	ret

;-----------------------------------------------------------------------------
milliwait:			;Wartezeit: r17:r16 msec, min 1 max 999
	push r18
	push r19
	push r16
	push r17
	cli
	mov r18, r10
	mov r19, r11		;r19r18 = sollzeit = aktuelle Millisec
	sei
	add r18, r16
	adc r19, r17		;sollzeit += Wartezeit
	ldi r16, 0xE8		;auf 1000 testen
	cpi r19, 0x03		;oberes Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
	brne subtra		;sollzeit>1000 --> subtrahiere 1000
	cp r18, r16		;unteres Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
subtra:	sub r18, r16
	sbci r19, 0x03		;wenn sollzeit>=1000 dann sollzeit-=1000
loop1:	cp r18, r10
	brne loop1
	cp r19, r11		;auf sollzeit warten
	brne loop1
	pop r17
	pop r16
pop1918ret:
	pop r19
	pop r18
	ret
